/*
 * Traditional system functions for memory management.
 *
 * Copyright (C) 2019 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/sys/types.h>

#include <errno.h>
#include <malloc.h>
#include <stdlib.h>     /* labs */

#include "mem_ipc.h"
#include "sys_memory.h"



/* NOTE: This should use the reentrancy state and prevent concurrent access. */

/* The current and previous limit of the data segment. */

static void *program_break = 0;
static void *program_break_previous = 0;

/* Available memory already allocated. */

const unsigned long initial_memory_size = L4_PAGESIZE;
unsigned long memory_available = 0;



/* Return the "program break" or address of the first location after the end of
   the data segment. If increment is non-zero, allocate more data space and
   provide the address before allocation as the result. */

void *_sbrk_r(struct _reent *reent, intptr_t increment)
{
  l4_cap_idx_t ds;
  void *addr;
  unsigned long size;

  /* For a zero increment, obtain the current limit. */

  if (!increment)
  {
    /* Without any recorded address, obtain some memory first in order to provide
       the initial address. */

    if (!program_break)
    {
      if (ipc_allocate(initial_memory_size, &program_break_previous, &ds))
      {
        reent->_errno = ENOMEM;
        return (void *) -1;
      }

      if (ipc_dataspace_size(ds, &memory_available))
      {
        reent->_errno = ENOMEM;
        return (void *) -1;
      }

      /* The actual limit is concealed from the caller. */

      program_break = program_break_previous;
    }
  }

  /* For a non-zero increment, obtain a new limit, using any available memory
     from initialisation. */

  /* NOTE: For now, ignore negative increments. */

  else if (increment > 0)
  {
    program_break_previous = program_break;

    /* Consume available pre-allocated memory if possible. */

    size = labs(increment);

    if (memory_available >= size)
      memory_available -= size;

    /* Without any available pre-allocated memory, obtain a new dataspace and
       advance the address. */

    else
    {
      if (ipc_allocate(size - memory_available, &addr, &ds))
      {
        reent->_errno = ENOMEM;
        return (void *) -1;
      }

      if (ipc_dataspace_size(ds, &size))
      {
        reent->_errno = ENOMEM;
        return (void *) -1;
      }

      /* If allocation does not produce an adjacent block, just return the
         address of the newly-allocated memory, because the caller will be
         expecting contiguous memory. Any remaining pre-allocated memory will be
         abandoned. */

      if (addr != program_break + memory_available)
        program_break_previous = addr;

      memory_available = 0;
    }

    /* Calculate the new limit using whichever size was employed. */

    program_break = program_break_previous + size;
  }

  /* Return the previous limit. */

  return program_break_previous;
}
